@@ -2,7 +2,7 @@
 /*
     no-ip.com dynamic IP update client for Linux
 
-   Copyright 2000-2002 Free Software Foundation, Inc.
+   Copyright 2000-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -79,6 +79,26 @@
 	+	November 2003 (johna)
 	+	added <CR> into all http requests along with <LF>
 	+	added SIGCHLD handler to reap zombies
+
+	+	January 2004 (johna)
+	+	added location logic and revamped XML parsing
+	+	added User-Agent field to settings.php request
+	+	changed to version 2.1
+
+	+	January 2004 (johna)	version 2.1.1
+	+	added -u, -p and -x options for LR101 project
+
+	+	April 2004 (johna	version 2.1.2
+	+	removed -Y in make install rule
+
+	+	August 2005 (johna)
+	+	added shm dump code for debugging broken libraries
+	+	added -z flag to invoke shm dump code
+
+	+	February 2006 (johna)
+	+	added code to handle new pedantic version of gcc
+	+	made signed/unsigned char assignments explicit
+
 */			
 /////////////////////////////////////////////////////////////////////////////                                            
 
@@ -108,6 +128,7 @@
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #include <pwd.h>
+#include <time.h>
 
 #ifdef bsd_with_getifaddrs
 #include <ifaddrs.h>
@@ -138,15 +159,18 @@
 #define READ_TIMEOUT		90
 #define WRITE_TIMEOUT		60
 #define CONNECT_TIMEOUT		60
+#define FORCE_INTERVAL		(1440 * 30)		// 30 days in muinutes
 
-#define CLIENT_IP_PORT		8245
 #define IPLEN			16
 #define LINELEN 	        256
 #define BIGBUFLEN		16384
 
-#define VERSION			"2.0.15"
+
 #define NOIP_NAME		"dynupdate.no-ip.com"
-#define USER_AGENT		"User-Agent: Linux DUC "VERSION
+#define CLIENT_IP_PORT		8245
+
+#define VERSION			"2.1.3"
+#define USER_AGENT		"User-Agent: Linux-DUC/"VERSION
 #define SETTING_SCRIPT		"settings.php?"
 #define USTRNG			"username="
 #define PWDSTRNG		"&pass="
@@ -159,20 +183,23 @@
   #define UPDATE_SCRIPT		"ducupdate.php"
 
 #ifdef DEBUG
-  #define OPTCHARS		"CYU:Fc:dD:hSMi:K:I:"
+  #define OPTCHARS		"CYU:Fc:dD:hp:u:x:SMi:K:I:z"
 #else
-  #define OPTCHARS		"CYU:Fc:hSMi:K:I:"
+  #define OPTCHARS		"CYU:Fc:hp:u:x:SMi:K:I:z"
 #endif
 #define ARGC			1
-#define ARGF			2
-#define ARGY			4
-#define ARGU			8
-#define ARGI			16
-#define ARGD			32
-#define ARGS			64
-#define ARGM			128
-#define ARGK			256
-#define ARGi			512
+#define ARGF			(1<<1)
+#define ARGY			(1<<2)
+#define ARGU			(1<<3)
+#define ARGI			(1<<4)
+#define ARGu			(1<<5)
+#define ARGp			(1<<6)
+#define ARGx			(1<<7)
+#define ARGD			(1<<8)
+#define ARGS			(1<<9)
+#define ARGM			(1<<10)
+#define ARGK			(1<<11)
+#define ARGi			(1<<12)
 
 #define NODNSGROUP		"@@NO_GROUP@@"
 #define HOST			1
@@ -274,6 +301,10 @@
 #define CMSG99	"\nThis client has expired.  Please go to http://www.no-ip.com/downloads.php"  
 #define CMSG99a	"to get our most recent version.\n"  
 
+#define CMSG100 \
+"Error! -C option can only be used with -F -Y -U -I -c -u -p -x options."
+#define CMSG101	" Both -u and -p options must be used together."
+
 int	debug			= 	0;
 int	timed_out		=	0;
 int	background		=	1;	// new default
@@ -295,6 +326,10 @@ int	multiple_instances	=	0;
 int	debug_toggle		=	0;
 int	kill_proc		=	0;
 int	reqnum			=	0;
+int	prompt_for_executable	=	1;
+int	shm_dump_active		=	0;
+int	Force_Update		=	0;
+int	dummy			=	0;
 void	*shmaddr		=	NULL;
 char	*program		=	NULL;
 char	*ourname		=	NULL;
@@ -303,6 +338,9 @@ char	*config_filename	=	NULL;
 char	*request		=	NULL;
 char	*pid_path		=	NULL;
 char	*execpath		=	NULL;
+char	*supplied_username	=	NULL;
+char	*supplied_password	=	NULL;
+char	*supplied_executable	=	NULL;
 char	IPaddress[IPLEN];
 char	login[LINELEN];
 char	password[LINELEN];
@@ -343,12 +381,10 @@ struct SETTINGS {
         char    *keyword;
         int     (*func)(char *);
 } settings[] = {
-        {       2,      "</",                   ignore  },
-        {       5,      "<?xml",                ignore  },
-        {       15,     "<noip_host_list",      ignore  },
         {       7,      "<domain",              domains },
         {       5,      "<host",                hosts   },
-        {       0,      "",                     xmlerr  },
+        {       1,      "<",                    ignore  }, // other objects
+        {       0,      "",                     xmlerr  }  // error msgs
 };
 
 struct GROUPS {
@@ -399,8 +435,9 @@ void	process_options(int argc, char *argv[]);
 void	alarm_handler(int signum);
 void	sigchld_handler(int signum);
 void	exit_handler(int signum);
-volatile void	wake_handler(int signum);
+void	wake_handler(int signum);
 void	save_IP();
+void	dump_shm(struct shmid_ds *d);
 int	get_shm_info();
 void	get_our_visible_IPaddr(char *dest);
 void	getip(char *dest, char *device);
@@ -417,29 +454,32 @@ void	display_current_config();
 void	display_one_config(char *req, int interval, 
 			int nat, char *device, char *expath);
 int	parse_config();
-ushort	chksum(unsigned char *buf, unsigned int size);
+ushort	chksum(char *buf, unsigned int size);
 void	dump_buffer(int x);
 int	dynamic_update();
 int	handle_dynup_error(int error_code);
 int	handle_config_error(int err_code);
-void	*Malloc(unsigned int size);
+void	*Malloc(int size);
+int	get_xml_field(char *name, char *line, char *dest);
 void	add_to_list(char *gnm, struct NAMES *ns);
 int	yesno(char *fmt, ...);
 int	add_to_request(int kind, char *p);
 int	get_update_selection(int tgrp, int thst);
 int	GetNextLine(char *dest);
-void	url_encode(unsigned char *in, unsigned char *out);
+void	url_encode(char *in, char *out);
 void	get_credentials(char *l, char *p);
 void	get_device_name(char *d);
 void	autoconf();
-int     bencode(const unsigned char *s, unsigned char *dst);
-int     bdecode(unsigned char *in, unsigned char *out);
+int     bencode(const char *s, char *dst);
+int     bdecode(char *in, char *out);
 void	Msg(char *fmt, ...);
 ///////////////////////////////////////////////////////////////////////////
 void Usage()
 {
 	fprintf(stderr,  "\nUSAGE: %s ", program);
-	fprintf(stderr,  "[ -C [ -F][ -Y][ -U #min]][ -c file]\n\t");
+	fprintf(stderr,  "[ -C [ -F][ -Y][ -U #min]\n\t");
+	fprintf(stderr,  "[ -u username][ -p password][ -x progname]]\n\t");
+	fprintf(stderr,  "[ -c file]");
 #ifdef DEBUG
         fprintf(stderr, "[ -d][ -D pid]");
 #endif                                                                          
@@ -449,6 +489,9 @@ void Usage()
 	fprintf(stderr, "         -F               force NAT off\n");
 	fprintf(stderr, "         -Y               select all hosts/groups\n");
 	fprintf(stderr, "         -U minutes       set update interval\n");
+	fprintf(stderr, "         -u username      use supplied username\n");
+	fprintf(stderr, "         -p password      use supplied password\n");
+	fprintf(stderr, "         -x executable    use supplied executable\n");
 	fprintf(stderr, "         -c config_file   use alternate data path\n");
 #ifdef DEBUG
         fprintf(stderr, "         -d               increase debug verbosity\n");
@@ -459,6 +502,7 @@ void Usage()
 	fprintf(stderr, "         -S               show configuration data\n");
 	fprintf(stderr, "         -M               permit multiple instances\n");
 	fprintf(stderr, "         -K processID     terminate instance PID\n");
+	fprintf(stderr, "         -z               activate shm dump code\n");
 	fprintf(stderr, "         -h               help (this text)\n");
 }
 ///////////////////////////////////////////////////////////////////////////
@@ -593,6 +637,7 @@ void process_options(int argc, char *argv[])
 		case 'd':
 			debug++;
 			log2syslog = -1;
+			shm_dump_active++;
 			strcat(saved_args, " -d");
 			break;
 		case 'D':
@@ -600,6 +645,18 @@ void process_options(int argc, char *argv[])
 			have_args |= ARGD;
 			break;
 #endif
+		case 'u':
+			supplied_username = optarg;
+			have_args |= ARGu;
+			break;
+		case 'p':
+			supplied_password = optarg;
+			have_args |= ARGp;
+			break;
+		case 'x':
+			supplied_executable = optarg;
+			have_args |= ARGx;
+			break;
 		case 'h':
 			Usage();
 			exit(0);
@@ -628,16 +685,31 @@ void process_options(int argc, char *argv[])
 			strcat(saved_args, optarg);
 			have_args |= ARGI;
 			break;
-
+		case 'z':
+			shm_dump_active++;
+			break;
 		default:
 			Usage();
 			exit(0);
 		}
 	}
-	if (needs_conf && (have_args > (ARGC|ARGF|ARGY|ARGU|ARGI))){
-	    Msg("Error! -C option can only be used with -F -Y -U -I -c options.");
+	if (needs_conf && (have_args > ((ARGx * 2) - 1))){
+	    Msg(CMSG100);
 	    exit(1);
 	}
+	if (have_args & ARGu) {
+	    if (!(have_args & ARGp)) {
+		Msg(CMSG101);
+		exit(1);
+	    }
+	}
+	if (have_args & ARGp) {
+	    if (!(have_args & ARGu)) {
+		Msg(CMSG101);
+		exit(1);
+	    }
+	    prompt_for_executable = 0;		// don't prompt in auto mode
+	}
 	if (debug_toggle && (have_args != ARGD)){
 	    Msg("Error! -D option can't be used with any other options.");
 	    exit(1);
@@ -675,8 +747,9 @@ void exit_handler(int signum)	//entered on SIGINT,  SIGTERM & SIGQUIT
 	// don't reset handler -- exit on second signal
 }
 ///////////////////////////////////////////////////////////////////////////
-volatile void wake_handler(int signum)	//entered on SIGUSR1
+void wake_handler(int signum)	//entered on SIGUSR1
 {			// used to wake sleeping process only
+	dummy++;
 }
 ///////////////////////////////////////////////////////////////////////////
 void save_IP()
@@ -768,6 +841,26 @@ int config_file_inuse()
 	return retval;
 }
 /////////////////////////////////////////////////////////////////////////
+void dump_shm(struct shmid_ds *d)
+{
+	if (shm_dump_active) {
+	    fprintf(stderr, "\nShared memory info:\n");
+	    fprintf(stderr, "\tsize of struct shmid_ds is %d\n", sizeof(struct shmid_ds));
+	    fprintf(stderr, "\tmode %o\n", d->shm_perm.mode & 0x1ff);
+	    fprintf(stderr, "\tuid %d, gid %d\n", 
+					d->shm_perm.uid, d->shm_perm.gid);
+	    fprintf(stderr, "\tcreated by uid %d, gid %d\n", 
+					d->shm_perm.cuid, d->shm_perm.cgid);
+	    fprintf(stderr, "\tsegment size = %d\n", d->shm_segsz);
+	    fprintf(stderr, "\tlast attach time = %s",   ctime(&d->shm_atime));
+	    fprintf(stderr, "\tlast detach time = %s",   ctime(&d->shm_dtime));
+	    fprintf(stderr, "\tcreation time = %s",   ctime(&d->shm_ctime));
+	    fprintf(stderr, "\tnumber of attached procs = %ld\n", d->shm_nattch);
+	    fprintf(stderr, "\tcreation process id = %d\n", d->shm_cpid);
+	    fprintf(stderr, "\tlast user process id = %d\n", d->shm_lpid);
+	}
+}
+/////////////////////////////////////////////////////////////////////////
 int get_shm_info()
 {
 	int	i, flag;
@@ -786,14 +879,17 @@ int get_shm_info()
 	if ((ds.shm_nattch > 0) && (!multiple_instances)) { 
 	    Msg("One %s process is already active,", program);
 	    Msg("and the multiple instance flag (-M) is not set.");
+	    dump_shm(&ds);
 	    return FATALERR;
 	}
 	if (ds.shm_nattch >= MAX_INSTANCE) { 
 	    Msg("Too many %s processes active. Ending!", program);
+	    dump_shm(&ds);
 	    return FATALERR;
 	}
 	if ((shmaddr = shmat(shmid, 0, 0)) == (void *)-1) {
 	    Msg("Can't attach shared memory. (%s) Ending!", strerror(errno));
+	    dump_shm(&ds);
 	    return FATALERR;
 	}
 	shared = (struct SHARED *)shmaddr;
@@ -869,13 +965,16 @@ int run_as_background()
 		    if (my_instance->debug) 
 		        Msg("! Last_IP_Addr = %s, IP = %s",my_instance->Last_IP_Addr, IPaddress);
 #endif
-		    if (*IPaddress && strcmp(IPaddress, my_instance->Last_IP_Addr))  {
+		    if (*IPaddress && ((Force_Update == 0) ||
+			    strcmp(IPaddress, my_instance->Last_IP_Addr)))  {
 			if (dynamic_update() == SUCCESS) 
 			    strcpy(my_instance->Last_IP_Addr, IPaddress);
 			*IPaddress = 0;
 			if (connect_fail)
 			    delay += 300;
 		    }
+		    if (--Force_Update < 0)
+			Force_Update = FORCE_INTERVAL / my_instance->interval;
 		    if (background)	// signal may have reset this!
 			Sleep(delay);
 		}
@@ -1029,11 +1128,13 @@ char *despace(char *p)
 	return p;
 }
 ///////////////////////////////////////////////////////////////////////////
-ushort chksum(unsigned char *buf, unsigned int size)
+ushort chksum(char *p_buf, unsigned int size)
 {
 	int	i;
 	ushort	sum;
+	unsigned char *buf;
 
+	buf = (unsigned char *)p_buf;
 	sum = 0;
 	for (i=0; i<size; i++)
             sum += *buf++;
@@ -1355,7 +1456,7 @@ void dump_buffer(int x)
 int dynamic_update()
 {
 	int	i, x, is_group, retval, response;
-	char	*p, *pos, tbuf[LINELEN], gname[LINELEN];
+	char	*p, *pos, tbuf[LINELEN], ipbuf[LINELEN], gname[LINELEN];
 
 	retval = SUCCESS;
 	if ((x = Connect(port_to_use)) != SUCCESS) {
@@ -1426,7 +1527,7 @@ int dynamic_update()
 		    Msg("Can't fork (%s) to run %s", strerror(errno), buffer);
 		    exit(1);
 		case 0:			// child
-		    sprintf(tbuf, "%s", IPaddress);
+		    sprintf(ipbuf, "%s", IPaddress);
 		    p = strrchr(buffer, '/');
 		    if (p)
 			p++;
@@ -1434,15 +1535,16 @@ int dynamic_update()
 			p = buffer;
 #ifdef DEBUG
 		    if (my_instance ? my_instance->debug : debug)
-			Msg("Running %s %s %s", buffer, p, tbuf);
+			Msg("Running %s %s %s %s", buffer, p, ipbuf, ourname);
 #endif
-	    	    execl(buffer, p, tbuf, ourname, (char *)0);
+	    	    execl(buffer, p, ipbuf, ourname, (char *)0);
 		    Msg("execl %s failed (%s)", buffer, strerror(errno));
 		    exit(0);		// bad execute -- don't care!
 		default:		// parent
 		    break;
 	    }
 	}
+	Force_Update = -1;
 	return retval;
 }
 /////////////////////////////////////////////////////////////////////////////
@@ -1570,7 +1672,7 @@ int handle_config_error(int err_code)
 	return FATALERR;
 }
 //////////////////////////////////////////////////////////////////////////
-void *Malloc(unsigned int size)
+void *Malloc(int size)
 {
         void *x = malloc(size);
         if (x != NULL)
@@ -1581,51 +1683,44 @@ void *Malloc(unsigned int size)
 //////////////////////////////////////////////////////////////////////////
 int domains(char *p)
 {
-        char *q, *d;
+        int x;
 
-        q = strchr(p, '=') + 2;
-        d = dmn;
-        while (*q != '"')
-            *d++ = *q++;
-        *d = 0;
-//fprintf(stderr, "DOMAIN = %s\n", dmn);
+        x = get_xml_field("name", p, dmn);
         return 0;
 }
 //////////////////////////////////////////////////////////////////////////
 int hosts(char *p)
 {
-        char    *q, *d, *dm, *g;
         char    fqdn[LINELEN];
         char    gnm[LINELEN];
+        char    locn[LINELEN];
+        char    *d, *dm;
+        int     x, y, z;
 
-        q = strchr(p, '=') + 2;
-        d = fqdn;
-        while (*q != '"')
-            *d++ = *q++;
-        *d++ = '.';
-        dm = dmn;
-        while (*dm)
-            *d++ = *dm++;
-        *d = 0;
-	if (debug > 1)
-	    fprintf(stderr, "HOST = %s\n", fqdn);
-        d = (char *)Malloc(strlen(fqdn)+1);
-        strcpy(d, fqdn);
+        x = get_xml_field("name", p, fqdn);
+        d = &fqdn[x];
+        if (x > 0) {
+            *d++ = '.';                         // append domain name
+            dm = dmn;
+            while (*dm)
+                *d++ = *dm++;
+            *d = 0;
+        }
+        y = get_xml_field("group", p, gnm);
+        z = get_xml_field("location", p, locn);
+	if ((z > 0) && (x > 0)) {               // does location exist?
+	    *d++ = '@';                         // append @location
+	    memcpy(d, locn, z+1);               // copy the null too!
+	}       
+	d = (char *)Malloc(strlen(fqdn)+1);
+	strcpy(d, fqdn);			// may be empty string
         ns = (struct NAMES *)Malloc(sizeof(struct NAMES));
         ns->link = NULL;
         ns->use = 0;
-        ns->fqdn = d;
-        while (strncmp(q, "group=", 6) != 0) {
-            q++;
-        }
-        q += 7;
-        g = gnm;
-        while (*q != '"')
-            *g++ = *q++;
-        *g = 0;
+	ns->fqdn = d;				// insert fqdn into struct
+        if ((debug > 1) && *fqdn)
+            fprintf(stderr, "FQDN = %s\n", fqdn);
         add_to_list(gnm, ns);
-	if (debug > 1)
-	    fprintf(stderr, "GROUP = %s\n", gnm);
         return 0;
 }
 //////////////////////////////////////////////////////////////////////////
@@ -1651,6 +1746,34 @@ int ignore(char *p)
 {
         return 0;
 }
+/////////////////////////////////////////////////////////////////////////////
+int get_xml_field(char *name, char *line, char *dest)
+{
+        int     len;
+        char *p, *q;
+
+        len = strlen(name);
+        p = line;
+        *dest = 0;
+        while (1) {
+            p = strchr(p, '=');
+            if (p == NULL)
+                break;
+//fprintf(stderr, "gxf: p = %s\n", &p[-len]);
+            if (strncasecmp(&p[-len],name, len) == 0) {
+                q = p+2;
+                p = dest;
+                while (*q && (*q != '"'))
+                    *p++ = *q++;
+                *p = 0;
+                break;
+            }
+            p++;                                        // step past '='
+        }
+        if ((debug > 1) && *dest)
+            fprintf(stderr, "GET_XML_FIELD: %s = %s\n", name, dest);
+        return strlen(dest);
+}
 //////////////////////////////////////////////////////////////////////////
 void add_to_list(char *gnm, struct NAMES *ns)
 {
@@ -1811,18 +1934,24 @@ int GetNextLine(char *dest)
         while (*p) {
             *q++ = *p;
             offset++;
-            if (*p++ == '\n')  {
+            if (*p == '\n')  {
                 *q = 0;
 //fprintf(stderr, "LINE = %s", dest);
                 return 1;
-            }
+            } else
+		p++;
         }
+	if (q > dest)
+	    return 1;
         return 0;
 }
 ///////////////////////////////////////////////////////////////////////////
-void url_encode(unsigned char *in, unsigned char *out)
+void url_encode(char *p_in, char *p_out)
 {
-        unsigned char ch;
+        unsigned char ch, *in, *out;
+
+	in  = (unsigned char *)p_in;
+	out = (unsigned char *)p_out;
 
         while ((ch = *in++)) {
             switch(ch) {
@@ -1832,13 +1961,13 @@ void url_encode(unsigned char *in, unsigned char *out)
 		case '@':  case '[': case '\\': case ']': case '^':
                 case '`': case '{': case '|': case '}': case '~':
 			*out++ = '%';
-			sprintf(out, "%2.2x", ch);
+			sprintf((char *)out, "%2.2x", ch);
 			out += 2;
 			break;
                 default: 
 		       if ((ch & 0x80) || (ch < 0x20)) {
                            *out++ = '%';
-                           sprintf(out, "%2.2x", ch);
+                           sprintf((char *)out, "%2.2x", ch);
                            out += 2;
                        } else
                            *out++ = ch;
@@ -1853,6 +1982,11 @@ void get_credentials(char *l, char *p)
 	char	*x;
 	unsigned char	ch = 0;
 
+	if (supplied_username) {		// have both uname/passwd
+	    url_encode(supplied_username, l);
+	    url_encode(supplied_password, p);
+	    return;
+	}
 	fprintf(stderr, CMSG30);
 	fgets(answer, LINELEN, stdin);
 	answer[strlen(answer) - 1] = 0;		// remove newline
@@ -1890,11 +2024,12 @@ void get_credentials(char *l, char *p)
 	url_encode(answer, p);
 }
 /////////////////////////////////////////////////////////////////////////////
-int get_all_device_names(unsigned char *devs)
+int get_all_device_names(char *devs)
 {
 	int		devnum=0;
 	unsigned char	*p, *q, *dq;
 
+
 #ifdef bsd_with_getifaddrs
 
 	struct ifaddrs *ifap, *ifa;
@@ -1902,7 +2037,7 @@ int get_all_device_names(unsigned char *devs)
 	    perror("getifaddrs()");
 	    exit(1);
 	}
-	dq = devs;     // point at name list
+	dq = (unsigned char *)devs;     // point at name list
 	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
 		if (ifa->ifa_addr->sa_family == AF_LINK) {
 // FreeBSD doesn't define some of these!
@@ -1952,7 +2087,7 @@ int get_all_device_names(unsigned char *devs)
         }
 	ifrp = ibuf;
 	ifend = (struct ifreq *)((char *)ibuf + ifc.ifc_len);
-	dq = devs;     // point at name list
+	dq = (unsigned char *)devs;     // point at name list
 	for (; ifrp < ifend; ifrp = ifnext) {
 	    q = dq;     // add new name into list
 	    p = ifrp->ifr_name;
@@ -1993,10 +2128,10 @@ int get_all_device_names(unsigned char *devs)
 	    exit(1);
 	}
 	num_ifreq = Ifc.ifc_len / sizeof(struct ifreq);
-	dq = devs;     // add new name into list
+	dq = (unsigned char *)devs;     // add new name into list
 	for (pIfr=Ifc.ifc_req,i=0; i<num_ifreq; pIfr++,i++) {
 	    q = dq;     // add new name into list
-	    p = pIfr->ifr_name;
+	    p = (unsigned char *)pIfr->ifr_name;
 	    if (strncmp("lo", p, 2) == 0)	// ignore loopbacks
 		continue;
 	    if (strchr(p, ':') != NULL)		// ignore 'virtual' devices
@@ -2022,13 +2157,13 @@ void get_device_name(char *d)
 	int	i, devnum;
 	unsigned char devs[MAX_NET_DEVS][LINELEN];
 
-	devnum = get_all_device_names((unsigned char *)&devs);
+	devnum = get_all_device_names((char *)&devs);
 	switch (devnum) {
 	    case 0:
 		Msg(CMSG19);
 		exit(1);
 	    case 1: 		// most common case
-		strcpy(d, devs[0]);
+		strcpy(d, (char *)devs[0]);
 		break;
 	    default:		// firewall machine
 		Msg(CMSG20);
@@ -2040,7 +2175,7 @@ void get_device_name(char *d)
 		    fgets(answer, LINELEN, stdin);
 		    i = atoi(answer);
 		} while ((i < 0) || (i >= devnum));
-		strcpy(d, devs[i]);
+		strcpy(d, (char *)devs[i]);
 		break;
 	}
 }
@@ -2117,8 +2252,8 @@ void autoconf()
 	}
 	sprintf(line, "%s%s%s%s", USTRNG, login, PWDSTRNG, password);
 	bencode(line, encline);
-        sprintf(buffer, "GET http://%s/%s%s%s\r\n\r\n", 
-                		NOIP_NAME, SETTING_SCRIPT, REQUEST, encline);
+        sprintf(buffer, "GET http://%s/%s%s%s HTTP/1.0\r\n%s\r\n\r\n", 
+       		NOIP_NAME, SETTING_SCRIPT, REQUEST, encline, USER_AGENT);
         if ((x = converse_with_web_server()) <= 0) {
             handle_dynup_error(x);
 	    return;
@@ -2132,6 +2267,8 @@ void autoconf()
         groups->glink = NULL;
         groups->nlink = NULL;
 
+	for (x=0; x<6; x++)	// skip first 6 lines HTTP protocol stuff
+            GetNextLine(line);
         while (GetNextLine(line)) {
             p = line;
 	    s = settings;
@@ -2168,15 +2305,21 @@ void autoconf()
 	new_config->interval = update_cycle;
 	pos = &buffer[BIGBUFLEN / 2];
 	*pos = 0;
-	if (yesno(CMSG40)) {		// ask about pgm to run at update
-	    fprintf(stderr, CMSG40a);
-	    fgets(answer, LINELEN, stdin);
-	    answer[strlen(answer) -1] = 0;	// remove newline
-	    realpath(answer, buffer);
+	if (supplied_executable) {
+	    realpath(supplied_executable, buffer);
 	    bencode(buffer, pos);
 	    new_config->elength = strlen(pos) + 1;
 	} else {
-	    new_config->elength = 0;
+	    if (prompt_for_executable && yesno(CMSG40)) {// ask about pgm to run at update
+		fprintf(stderr, CMSG40a);
+		fgets(answer, LINELEN, stdin);
+		answer[strlen(answer) -1] = 0;	// remove newline
+		realpath(answer, buffer);
+		bencode(buffer, pos);
+		new_config->elength = strlen(pos) + 1;
+	    } else {
+		new_config->elength = 0;
+	    }
 	}
 	new_config->encrypt = ENCRYPT;
 	new_config->count = reqnum;
@@ -2201,15 +2344,20 @@ void autoconf()
 	return;
 }
 /////////////////////////////////////////////////////////////////////////
-int bencode(const unsigned char *s, unsigned char *dst)
+int bencode(const char *p_s, char *p_dst)
 {                         // http basic authorization encoding (base64)
 #if ENCRYPT
-        int n           = strlen(s);
-        int n3byt       = n / 3; 
-        int k           = n3byt * 3; 
-        int nrest       = n % 3;
-        int i           = 0;
-        int dstlen      = 0;
+        int n, n3byt, k, i, nrest, dstlen;
+	unsigned char *s, *dst;
+
+	s  	= (unsigned char *)p_s;
+	dst	= (unsigned char *)p_dst;
+        n  	= strlen(p_s);
+        n3byt   = n / 3; 
+        k       = n3byt * 3; 
+        nrest   = n % 3;
+        i       = 0;
+        dstlen  = 0;
 
         while (i < k) {
           dst[dstlen++] = EncodeTable[(( s[i]  & 0xFC)>>2)];
@@ -2234,17 +2382,22 @@ int bencode(const unsigned char *s, unsigned char *dst)
         dst[dstlen] = 0;
 	return dstlen;
 #else
-        strcpy(dst, s);
-        return strlen(s);
+        strcpy(p_dst, p_s);
+        return strlen(p_s);
 
 #endif
 }
 ////////////////////////////////////////////////////////////////////////////
-int bdecode(unsigned char *in, unsigned char *out)
+int bdecode(char *p_in, char *p_out)
 {
 #if ENCRYPT
-        int x = strlen(in);
+	unsigned char *in, *out;
         unsigned char *p, *q, d1, d2, d3, d4;
+        int x;
+
+	in  = (unsigned char *)p_in;
+	out = (unsigned char *)p_out;
+        x = strlen(p_in);
 
         p = q = &in[x];
         while (x % B64MOD) {      // pad to a multiple of four (if malformed)
@@ -2268,9 +2421,8 @@ int bdecode(unsigned char *in, unsigned char *out)
         } while (x > 0);
         *out = 0;
 	*q = 0;					// replace original null
-        return 0;
 #else
-	strcpy(out, in);
+	strcpy(p_out, p_in);
 #endif
 	return 0;
 }
